# pylint: disable=missing-docstring,too-many-lines
from unittest.mock import MagicMock, patch

from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError

from resolwe.flow.models import Collection, Data, DescriptorSchema, Entity, Process
from resolwe.flow.models.utils import validate_data_object, validate_schema
from resolwe.test import TestCase
from resolwe.test.utils import create_data_location


class ValidationTest(TestCase):
    def setUp(self):
        super().setUp()

        self.user = get_user_model().objects.create(username="test_user")

    def test_validating_data_object(self):
        """Diferent validations are performed depending on status"""
        proc = Process.objects.create(
            name="Test process",
            contributor=self.user,
            input_schema=[
                {"name": "value", "type": "basic:integer:", "required": True}
            ],
            output_schema=[
                {"name": "result", "type": "basic:string:", "required": True}
            ],
        )

        data = {
            "name": "Test data",
            "contributor": self.user,
            "process": proc,
        }

        with self.assertRaisesRegex(ValidationError, '"value" not given'):
            validate_data_object(Data.objects.create(input={}, **data))

        with self.assertRaisesRegex(ValidationError, "Required fields .* not given"):
            validate_data_object(Data.objects.create(input={}, **data))

        d = Data.objects.create(input={"value": 42}, **data)

        d.status = Data.STATUS_DONE
        with self.assertRaisesRegex(ValidationError, '"result" not given'):
            d.save()
            validate_data_object(d)

        d.output = {"result": "forty-two"}
        d.save()
        validate_data_object(d)

    def test_validate_data_descriptor(self):
        proc = Process.objects.create(name="Test process", contributor=self.user)
        descriptor_schema = DescriptorSchema.objects.create(
            name="Test descriptor schema",
            contributor=self.user,
            schema=[{"name": "description", "type": "basic:string:", "required": True}],
        )

        data = Data.objects.create(
            name="Test descriptor",
            contributor=self.user,
            process=proc,
            descriptor={},
            descriptor_schema=descriptor_schema,
        )
        self.assertEqual(data.descriptor_dirty, True)

        data.descriptor = {"description": "some value"}
        data.save()
        self.assertEqual(data.descriptor_dirty, False)

        data.descriptor = {}
        data.save()
        self.assertEqual(data.descriptor_dirty, True)

        data.descriptor = {"description": 42}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            data.save()

    def test_validate_collection_descriptor(self):
        descriptor_schema = DescriptorSchema.objects.create(
            name="Test descriptor schema",
            contributor=self.user,
            schema=[{"name": "description", "type": "basic:string:", "required": True}],
        )

        collection = Collection.objects.create(
            name="Test descriptor",
            contributor=self.user,
            descriptor_schema=descriptor_schema,
        )
        self.assertEqual(collection.descriptor_dirty, True)

        collection.descriptor = {"description": "some value"}
        collection.save()
        self.assertEqual(collection.descriptor_dirty, False)

        collection.descriptor = {}
        collection.save()
        self.assertEqual(collection.descriptor_dirty, True)

        collection.descriptor = {"description": 42}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            collection.save()

    def test_validate_entity_descriptor(self):
        descriptor_schema = DescriptorSchema.objects.create(
            name="Test descriptor schema",
            contributor=self.user,
            schema=[{"name": "description", "type": "basic:string:", "required": True}],
        )

        entity = Entity.objects.create(
            name="Test descriptor",
            contributor=self.user,
            descriptor_schema=descriptor_schema,
        )
        self.assertEqual(entity.descriptor_dirty, True)

        entity.descriptor = {"description": "some value"}
        entity.save()
        self.assertEqual(entity.descriptor_dirty, False)

        entity.descriptor = {}
        entity.save()
        self.assertEqual(entity.descriptor_dirty, True)

        entity.descriptor = {"description": 42}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            entity.save()

    def test_referenced_storage(self):
        proc = Process.objects.create(
            name="Test process",
            contributor=self.user,
            output_schema=[
                {"name": "big_result", "type": "basic:json:", "required": True}
            ],
        )

        data = {
            "name": "Test data",
            "contributor": self.user,
            "process": proc,
        }

        d = Data.objects.create(**data)
        d.status = Data.STATUS_DONE

        # `Data` object with referenced non-existing `Storage`
        d.output = {"big_result": 245}
        with self.assertRaisesRegex(ValidationError, "`Storage` object does not exist"):
            d.save()
            validate_data_object(d)

        d.storages.create(name="storage", contributor=self.user, json={"value": 42})
        self.assertEqual(d.storages.count(), 1)
        d.output = {"big_result": d.storages.first().id}
        d.save()
        validate_data_object(d)

    def test_referenced_data(self):
        proc1 = Process.objects.create(
            name="Referenced process",
            contributor=self.user,
            type="data:referenced:object:",
        )
        proc2 = Process.objects.create(
            name="Test process",
            contributor=self.user,
            input_schema=[{"name": "data_object", "type": "data:referenced:object:"}],
        )
        d = Data.objects.create(
            name="Referenced object", contributor=self.user, process=proc1
        )
        validate_data_object(d)

        data = {
            "name": "Test data",
            "contributor": self.user,
            "process": proc2,
            "input": {"data_object": d.pk},
        }

        validate_data_object(Data.objects.create(**data))

        # less specific type
        proc2.input_schema = [{"name": "data_object", "type": "data:referenced:"}]
        validate_data_object(Data.objects.create(**data))

        # wrong type
        proc2.input_schema = [{"name": "data_object", "type": "data:wrong:type:"}]
        with self.assertRaisesRegex(
            ValidationError, "Data object of type .* is required"
        ):
            validate_data_object(Data.objects.create(**data))

        # Non-existing referenced `Data` object on create.
        data["input"] = {"data_object": 631}
        with self.assertRaisesRegex(ValidationError, "`Data` object does not exist"):
            Data.objects.create(**data)

        # Non-existing `Data` object when status is set to `DONE` should not
        # raise exception. First make sure the object is created.
        proc2.input_schema = [{"name": "data_object", "type": "data:referenced:"}]
        data["input"] = {"data_object": d.pk}
        d = Data.objects.create(**data)
        data["input"] = {"data_object": 631}
        # Now validate with missing input.
        validate_data_object(d, skip_missing_data=True)

    def test_delete_input(self):
        proc1 = Process.objects.create(
            name="Referenced process",
            contributor=self.user,
            type="data:referenced:object:",
        )
        proc2 = Process.objects.create(
            name="Test process",
            contributor=self.user,
            input_schema=[{"name": "data_object", "type": "data:referenced:object:"}],
        )
        data1 = Data.objects.create(
            name="Referenced object", contributor=self.user, process=proc1
        )
        data2 = Data.objects.create(
            name="Test data",
            contributor=self.user,
            process=proc2,
            input={"data_object": data1.pk},
        )
        validate_data_object(data1)
        validate_data_object(data2)

        data1.delete()
        data2.name = "New name"
        data2.save()
        validate_data_object(data2, skip_missing_data=True)


class ValidationUnitTest(TestCase):
    def test_required(self):
        schema = [
            {"name": "value", "type": "basic:integer:", "required": True},
            {
                "name": "description",
                "type": "basic:string:",
            },  # implicit `required=True`
            {"name": "comment", "type": "basic:string:", "required": False},
        ]

        instance = {"description": "test"}
        with self.assertRaisesRegex(ValidationError, '"value" not given.'):
            validate_schema(instance, schema)

        instance = {"value": 42}
        with self.assertRaisesRegex(ValidationError, '"description" not given.'):
            validate_schema(instance, schema)

        instance = {"value": 42, "description": "universal answer"}
        validate_schema(instance, schema)

        instance = {"value": 42, "description": "universal answer", "comment": None}
        validate_schema(instance, schema)

        instance = {"value": 42, "description": "test", "comment": "Lorem ipsum"}
        validate_schema(instance, schema)

        instance = {}
        validate_schema(instance, schema, test_required=False)

    def test_choices(self):
        schema = [
            {
                "name": "value",
                "type": "basic:integer:",
                "choices": [{"value": 7, "label": "7"}, {"value": 13, "label": "13"}],
            },
        ]

        instance = {"value": 7}
        validate_schema(instance, schema)

        instance = {"value": 8}
        error_msg = "Value of field 'value' must match one of predefined choices. Current value: 8"
        with self.assertRaisesRegex(ValidationError, error_msg):
            validate_schema(instance, schema)

        schema = [
            {
                "name": "value",
                "type": "basic:integer:",
                "choices": [{"value": 7, "label": "7"}, {"value": 13, "label": "13"}],
                "allow_custom_choice": True,
            },
        ]

        instance = {"value": 7}
        validate_schema(instance, schema)

        instance = {"value": 8}
        validate_schema(instance, schema)

    def test_range(self):
        schema = [
            {
                "name": "value",
                "type": "basic:integer:",
                "range": [0, 5],
            },
        ]

        instance = {"value": 5}
        validate_schema(instance, schema)
        instance = {"value": 0}
        validate_schema(instance, schema)
        instance = {"value": 3}
        validate_schema(instance, schema)

        instance = {"value": 7}
        error_msg = (
            "Value of field 'value' is out of range. It should be between 0 and 5."
        )
        with self.assertRaisesRegex(ValidationError, error_msg):
            validate_schema(instance, schema)

        schema = [
            {
                "name": "value",
                "type": "basic:decimal:",
                "range": [0, 5],
            },
        ]

        instance = {"value": 7}
        error_msg = (
            "Value of field 'value' is out of range. It should be between 0 and 5."
        )
        with self.assertRaisesRegex(ValidationError, error_msg):
            validate_schema(instance, schema)

        schema = [
            {
                "name": "value",
                "type": "list:basic:integer:",
                "range": [0, 5],
            },
        ]

        instance = {"value": [0, 3, 5]}
        validate_schema(instance, schema)

        instance = {"value": [0, 3, 7]}
        error_msg = (
            "Value of field 'value' is out of range. It should be between 0 and 5."
        )
        with self.assertRaisesRegex(ValidationError, error_msg):
            validate_schema(instance, schema)

        schema = [
            {
                "name": "value",
                "type": "list:basic:decimal:",
                "range": [0, 5],
            },
        ]

        instance = {"value": [0, 3, 7]}
        error_msg = (
            "Value of field 'value' is out of range. It should be between 0 and 5."
        )
        with self.assertRaisesRegex(ValidationError, error_msg):
            validate_schema(instance, schema)

    def test_missing_in_schema(self):
        schema = [{"name": "result", "type": "basic:integer:", "required": False}]

        instance = {"res": 42}
        with self.assertRaisesRegex(ValidationError, r"\(res\) missing in schema"):
            validate_schema(instance, schema)

    def test_file_prefix(self):
        schema = [
            {"name": "result", "type": "basic:file:"},
        ]
        path_instance = MagicMock()
        instance = {"result": {"file": "result.txt"}}

        data_location = create_data_location(subpath="1")

        with patch("resolwe.flow.models.utils.validation.Path") as path_mock:
            path_mock.return_value = path_instance
            validate_schema(instance, schema)
            path_instance.exists = MagicMock(return_value=True)
            self.assertEqual(path_instance.is_file.call_count, 0)

        # missing file
        with patch("resolwe.flow.models.utils.validation.Path") as path_mock:
            path_mock.return_value = path_instance
            path_instance.exists = MagicMock(return_value=False)
            path_instance.__truediv__ = MagicMock(return_value=path_instance)
            with self.assertRaisesRegex(
                ValidationError, "Referenced path .* does not exist."
            ):
                validate_schema(instance, schema, data_location=data_location)
            self.assertEqual(path_instance.exists.call_count, 1)

        # File is not a normal file
        with patch("resolwe.flow.models.utils.validation.Path") as path_mock:
            path_mock.return_value = path_instance
            path_instance.exists = MagicMock(return_value=True)
            path_instance.is_file = MagicMock(return_value=False)
            path_instance.__truediv__ = MagicMock(return_value=path_instance)
            with self.assertRaisesRegex(
                ValidationError, "Referenced path .* is not a file."
            ):
                validate_schema(instance, schema, data_location=data_location)
            self.assertEqual(path_instance.is_file.call_count, 1)

        with patch("resolwe.flow.models.utils.validation.Path") as path_mock:
            path_mock.return_value = path_instance
            path_instance.is_file = MagicMock(return_value=True)
            path_instance.__truediv__ = MagicMock(return_value=path_instance)
            validate_schema(instance, schema, data_location=data_location)
            self.assertEqual(path_instance.is_file.call_count, 1)

        instance = {
            "result": {"file": "result.txt", "refs": ["user1.txt", "user2.txt"]}
        }

        with patch("resolwe.flow.models.utils.validation.Path") as path_mock:
            path_mock.return_value = path_instance
            path_instance.is_file = MagicMock(return_value=True)
            path_instance.__truediv__ = MagicMock(return_value=path_instance)
            validate_schema(instance, schema, data_location=data_location)
            self.assertEqual(path_instance.is_file.call_count, 3)

        # missing second `refs` file
        with patch("resolwe.flow.models.utils.validation.Path") as path_mock:
            path_mock.return_value = path_instance
            path_instance.is_file = MagicMock(side_effect=[True, True, False])
            path_instance.is_dir = MagicMock(return_value=False)
            path_instance.__truediv__ = MagicMock(return_value=path_instance)
            message = "Path referenced in `refs` .* is neither a file or directory."
            with self.assertRaisesRegex(ValidationError, message):
                validate_schema(instance, schema, data_location=data_location)
            self.assertEqual(path_instance.is_file.call_count, 3)
            self.assertEqual(path_instance.is_dir.call_count, 1)

    def test_dir_prefix(self):
        schema = [
            {"name": "result", "type": "basic:dir:"},
        ]
        path_instance = MagicMock()
        instance = {"result": {"dir": "results"}}
        data_location = create_data_location(subpath="1")

        # dir validation is not called if `data_location` is not given
        with patch("resolwe.flow.models.utils.validation.Path") as path_mock:
            path_mock.return_value = path_instance
            validate_schema(instance, schema)
            self.assertEqual(path_instance.is_file.call_count, 0)
        path_instance.reset_mock()

        with patch("resolwe.flow.models.utils.validation.Path") as path_mock:
            path_mock.return_value = path_instance
            path_instance.exists = lambda: True
            path_instance.__truediv__ = MagicMock(return_value=path_instance)
            path_instance.is_dir = MagicMock(return_value=True)
            validate_schema(instance, schema, data_location=data_location)
            self.assertEqual(path_instance.is_dir.call_count, 1)
        path_instance.reset_mock()

        # missing path
        with patch("resolwe.flow.models.utils.validation.Path") as path_mock:
            path_mock.return_value = path_instance
            path_instance.exists = MagicMock(return_value=False)
            with self.assertRaisesRegex(
                ValidationError, "Referenced path .* does not exist."
            ):
                validate_schema(instance, schema, data_location=data_location)
            self.assertEqual(path_instance.exists.call_count, 1)
        path_instance.reset_mock()

        # not a dir
        with patch("resolwe.flow.models.utils.validation.Path") as path_mock:
            path_mock.return_value = path_instance
            path_instance.exists = MagicMock(return_value=True)
            path_instance.is_dir = MagicMock(return_value=False)
            with self.assertRaisesRegex(
                ValidationError, "Referenced path .* is not a directory"
            ):
                validate_schema(instance, schema, data_location=data_location)
            self.assertEqual(path_instance.is_dir.call_count, 1)
        path_instance.reset_mock()

        instance = {"result": {"dir": "results", "refs": ["file01.txt", "file02.txt"]}}

        with patch("resolwe.flow.models.utils.validation.Path") as path_mock:
            path_mock.return_value = path_instance
            path_instance.is_file = MagicMock(return_value=True)
            path_instance.is_dir = MagicMock(return_value=True)
            path_instance.__truediv__ = MagicMock(return_value=path_instance)

            validate_schema(instance, schema, data_location=data_location)
            self.assertEqual(path_instance.is_dir.call_count, 1)
            self.assertEqual(path_instance.is_file.call_count, 2)
        path_instance.reset_mock()

        # missing second `refs` file
        with patch("resolwe.flow.models.utils.validation.Path") as path_mock:
            path_mock.return_value = path_instance
            path_instance.is_file = MagicMock(side_effect=[True, False])
            path_instance.is_dir = MagicMock(side_effect=[True, False])
            path_instance.__truediv__ = MagicMock(return_value=path_instance)
            message = "Path referenced in `refs` .* is neither a file or directory"
            with self.assertRaisesRegex(ValidationError, message):
                validate_schema(instance, schema, data_location=data_location)
            self.assertEqual(path_instance.is_dir.call_count, 2)
            self.assertEqual(path_instance.is_file.call_count, 2)

    def test_string_field(self):
        schema = [
            {"name": "string", "type": "basic:string:"},
            {"name": "text", "type": "basic:text:"},
        ]

        instance = {"string": "Test string", "text": "Test text"}
        validate_schema(instance, schema)

        instance = {"string": 42, "text": 42}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

    def test_boolean_field(self):
        schema = [
            {"name": "true", "type": "basic:boolean:"},
            {"name": "false", "type": "basic:boolean:"},
        ]

        instance = {"true": True, "false": False}
        validate_schema(instance, schema)

        instance = {"true": "true", "false": "false"}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

        instance = {"true": 1, "false": 0}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

        instance = {"true": "foo", "false": "bar"}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

    def test_integer_field(self):
        schema = [
            {"name": "value", "type": "basic:integer:"},
        ]

        instance = {"value": 42}
        validate_schema(instance, schema)

        instance = {"value": 42.0}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

        instance = {"value": "forty-two"}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

    def test_decimal_field(self):
        schema = [
            {"name": "value", "type": "basic:decimal:"},
        ]

        instance = {"value": 42}
        validate_schema(instance, schema)

        instance = {"value": 42.0}
        validate_schema(instance, schema)

        instance = {"value": "forty-two"}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

    def test_date_field(self):
        schema = [
            {"name": "date", "type": "basic:date:"},
        ]

        instance = {"date": "2000-12-31"}
        validate_schema(instance, schema)

        instance = {"date": "2000/01/01"}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

        instance = {"date": "31 04 2000"}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

        instance = {"date": "21.06.2000"}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

        instance = {"date": "2000-1-1"}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

        instance = {"date": "2000 apr 8"}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

    def test_datetime_field(self):
        schema = [
            {"name": "date", "type": "basic:datetime:"},
        ]

        instance = {"date": "2000-06-21 00:00"}
        validate_schema(instance, schema)

        instance = {"date": "2000 06 21 24:00"}
        with self.assertRaises(ValidationError):
            validate_schema(instance, schema)

        instance = {"date": "2000/06/21 2:03"}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

        instance = {"date": "2000-06-21 2:3"}  # XXX: Is this ok?
        validate_schema(instance, schema)

        instance = {"date": "2000-06-21"}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

        instance = {"date": "2000-06-21 12pm"}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

    def test_data_field(self):
        schema = [{"name": "data_list", "type": "data:test:upload:"}]
        instance = {"data_list": 1}

        with patch("resolwe.flow.models.data.Data") as data_mock:
            value_mock = MagicMock(
                **{
                    "exists.return_value": True,
                    "first.return_value": {"process__type": "data:test:upload:"},
                }
            )
            filter_mock = MagicMock(**{"values.return_value": value_mock})
            data_mock.objects.filter.return_value = filter_mock

            validate_schema(instance, schema)

            self.assertEqual(value_mock.exists.call_count, 1)
            self.assertEqual(value_mock.first.call_count, 1)

        # subtype is OK
        with patch("resolwe.flow.models.data.Data") as data_mock:
            value_mock = MagicMock(
                **{
                    "exists.return_value": True,
                    "first.return_value": {
                        "process__type": "data:test:upload:subtype:"
                    },
                }
            )
            filter_mock = MagicMock(**{"values.return_value": value_mock})
            data_mock.objects.filter.return_value = filter_mock

            validate_schema(instance, schema)

            self.assertEqual(value_mock.exists.call_count, 1)
            self.assertEqual(value_mock.first.call_count, 1)

        # missing `Data` object
        with patch("resolwe.flow.models.data.Data") as data_mock:
            value_mock = MagicMock(
                **{
                    "exists.return_value": False,
                    "first.return_value": {"process__type": "data:test:upload:"},
                }
            )
            filter_mock = MagicMock(**{"values.return_value": value_mock})
            data_mock.objects.filter.return_value = filter_mock

            with self.assertRaisesRegex(
                ValidationError, "`Data` object does not exist"
            ):
                validate_schema(instance, schema)

            self.assertEqual(value_mock.exists.call_count, 1)
            self.assertEqual(value_mock.first.call_count, 0)

            validate_schema(instance, schema, skip_missing_data=True)

        # `Data` object of wrong type
        with patch("resolwe.flow.models.data.Data") as data_mock:
            value_mock = MagicMock(
                **{
                    "exists.return_value": True,
                    "first.return_value": {"process__type": "data:test:wrong:"},
                }
            )
            filter_mock = MagicMock(**{"values.return_value": value_mock})
            data_mock.objects.filter.return_value = filter_mock

            with self.assertRaisesRegex(
                ValidationError, "Data object of type .* is required"
            ):
                validate_schema(instance, schema)

            self.assertEqual(value_mock.exists.call_count, 1)
            self.assertEqual(value_mock.first.call_count, 1)

        # data `id` shouldn't be string
        instance = {"data_list": "1"}

        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

    def test_file_field(self):
        schema = [
            {"name": "result", "type": "basic:file:", "validate_regex": r"^.*\.txt$"},
        ]

        instance = {"result": {"file": "result_file.txt", "size": 13}}
        validate_schema(instance, schema)

        instance = {
            "result": {
                "file_temp": "12345",
                "file": "result_file.txt",
            }
        }
        validate_schema(instance, schema)

        instance = {
            "result": {
                "file_temp": "12345",
                "is_remote": True,
                "file": "result_file.txt",
            }
        }
        validate_schema(instance, schema)

        instance = {
            "result": {
                "file": "result_file.txt",
                "refs": ["01.txt", "02.txt"],
            }
        }
        validate_schema(instance, schema)

        # non-boolean `is_remote`
        instance = {
            "result": {
                "file_temp": "12345",
                "is_remote": "ftp",
                "file": "result_file.txt",
            }
        }
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

        # missing `file`
        instance = {
            "result": {
                "file_temp": "12345",
            }
        }
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

        # wrong file extension
        instance = {
            "result": {
                "file": "result_file.tar.gz",
            }
        }
        with self.assertRaisesRegex(
            ValidationError, "File name .* does not match regex"
        ):
            validate_schema(instance, schema)

    def test_html_file_field(self):
        schema = [
            {"name": "html_result", "type": "basic:file:html:"},
        ]

        instance = {
            "html_result": {"file": "index.htmls", "refs": ["some.js", "some.css"]}
        }
        validate_schema(instance, schema)

    def test_dir_field(self):
        schema = [
            {"name": "result", "type": "basic:dir:"},
        ]

        instance = {"result": {"dir": "results", "size": 32156}}
        validate_schema(instance, schema)

        instance = {"result": {"dir": "result", "refs": ["01.txt", "02.txt"]}}
        validate_schema(instance, schema)

        # missing `dir`
        instance = {
            "result": {
                "refs": ["01.txt", "02.txt"],
            }
        }
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

    def test_url_field(self):
        schema = [
            {"name": "url_view", "type": "basic:url:view:", "required": False},
            {"name": "url_download", "type": "basic:url:download:", "required": False},
            {"name": "url_link", "type": "basic:url:link:", "required": False},
        ]

        instance = {
            "url_view": {"url": "http://www.genialis.com"},
            "url_download": {"url": "http://www.genialis.com"},
            "url_link": {"url": "http://www.genialis.com"},
        }
        validate_schema(instance, schema)

        instance = {
            "url_view": {
                "url": "http://www.genialis.com",
                "name": "Genialis",
                "refs": ["http://www.genialis.com/jobs"],
            }
        }
        validate_schema(instance, schema)

        # wrong type
        schema = [
            {"name": "webpage", "type": "basic:url:"},
        ]
        instance = {"webpage": {"url": "http://www.genialis.com"}}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

    def test_json_field(self):
        schema = [{"name": "big_dict", "type": "basic:json:"}]

        # json not saved in `Storage`
        instance = {"big_dict": {"foo": "bar"}}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

        with patch("resolwe.flow.models.storage.Storage") as storage_mock:
            filter_mock = MagicMock()
            filter_mock.exists.return_value = True
            storage_mock.objects.filter.return_value = filter_mock

            instance = {"big_dict": 5}
            validate_schema(instance, schema)

            self.assertEqual(filter_mock.exists.call_count, 1)

        # non existing `Storage`
        with patch("resolwe.flow.models.storage.Storage") as storage_mock:
            filter_mock = MagicMock()
            filter_mock.exists.return_value = False
            storage_mock.objects.filter.return_value = filter_mock

            instance = {"big_dict": 5}
            with self.assertRaisesRegex(
                ValidationError, "`Storage` object does not exist"
            ):
                validate_schema(instance, schema)

            self.assertEqual(filter_mock.exists.call_count, 1)

    def test_list_string_field(self):
        schema = [{"name": "list", "type": "list:basic:string:"}]

        instance = {"list": ["foo", "bar"]}
        validate_schema(instance, schema)

        instance = {"list": []}
        validate_schema(instance, schema)

        instance = {"list": ""}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

        instance = {"list": "foo bar"}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

    def test_list_integer_field(self):
        schema = [
            {"name": "value", "type": "list:basic:integer:"},
        ]

        instance = {"value": [42, 43]}
        validate_schema(instance, schema)

        instance = {"value": 42}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

        instance = {"value": [42, 43.0]}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

        instance = {"value": [42, "43"]}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

    def test_list_data_field(self):
        schema = [{"name": "data_list", "type": "list:data:test:upload:"}]
        instance = {"data_list": [1, 3, 4]}

        with patch("resolwe.flow.models.data.Data") as data_mock:
            value_mock = MagicMock(
                **{
                    "exists.return_value": True,
                    "first.return_value": {"process__type": "data:test:upload:"},
                }
            )
            filter_mock = MagicMock(**{"values.return_value": value_mock})
            data_mock.objects.filter.return_value = filter_mock

            validate_schema(instance, schema)

            self.assertEqual(value_mock.exists.call_count, 3)
            self.assertEqual(value_mock.first.call_count, 3)

        # subtypes are OK
        with patch("resolwe.flow.models.data.Data") as data_mock:
            value_mock = MagicMock(
                **{
                    "exists.return_value": True,
                    "first.side_effect": [
                        {"process__type": "data:test:upload:subtype1:"},
                        {"process__type": "data:test:upload:"},
                        {"process__type": "data:test:upload:subtype2:"},
                    ],
                }
            )
            filter_mock = MagicMock(**{"values.return_value": value_mock})
            data_mock.objects.filter.return_value = filter_mock

            validate_schema(instance, schema)

            self.assertEqual(value_mock.exists.call_count, 3)
            self.assertEqual(value_mock.first.call_count, 3)

        # one object does not exist
        with patch("resolwe.flow.models.data.Data") as data_mock:
            value_mock = MagicMock(
                **{
                    "exists.side_effect": [True, False, True],
                    "first.return_value": {"process__type": "data:test:upload:"},
                }
            )
            filter_mock = MagicMock(**{"values.return_value": value_mock})
            data_mock.objects.filter.return_value = filter_mock

            with self.assertRaisesRegex(
                ValidationError, "`Data` object does not exist"
            ):
                validate_schema(instance, schema)

            self.assertEqual(value_mock.exists.call_count, 2)
            self.assertEqual(value_mock.first.call_count, 1)

        # one object of wrong type
        with patch("resolwe.flow.models.data.Data") as data_mock:
            value_mock = MagicMock(
                **{
                    "exists.return_value": True,
                    "first.side_effect": [
                        {"process__type": "data:test:upload:"},
                        {"process__type": "data:test:upload:"},
                        {"process__type": "data:test:wrong:"},
                    ],
                }
            )
            filter_mock = MagicMock(**{"values.return_value": value_mock})
            data_mock.objects.filter.return_value = filter_mock

            with self.assertRaisesRegex(
                ValidationError, "Data object of type .* is required"
            ):
                validate_schema(instance, schema)

            self.assertEqual(value_mock.exists.call_count, 3)
            self.assertEqual(value_mock.first.call_count, 3)

    def test_list_file_field(self):
        schema = [
            {
                "name": "result",
                "type": "list:basic:file:",
                "validate_regex": r"^.*\.txt$",
            },
        ]
        instance = {
            "result": [
                {"file": "result01.txt"},
                {"file": "result02.txt", "size": 14, "refs": ["results.tar.gz"]},
            ]
        }
        validate_schema(instance, schema)

        # wrong extension
        instance = {
            "result": [
                {"file": "result01.txt"},
                {"file": "result02.tar.gz"},
            ]
        }
        with self.assertRaisesRegex(
            ValidationError, "File name .* does not match regex"
        ):
            validate_schema(instance, schema)

    def test_list_dir_field(self):
        schema = [
            {"name": "result", "type": "list:basic:dir:"},
        ]

        instance = {
            "result": [
                {
                    "dir": "results01",
                    "size": 32156,
                    "refs": ["result01.txt", "result02.txt"],
                },
                {"dir": "results02"},
            ]
        }
        validate_schema(instance, schema)

        # missing `dir`
        instance = {
            "result": [
                {"size": 32156, "refs": ["result01.txt", "result02.txt"]},
            ]
        }
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

    def test_list_url_field(self):
        schema = [
            {"name": "webpage", "type": "list:basic:url:view:"},
        ]

        instance = {
            "webpage": [
                {
                    "url": "http://www.genialis.com",
                    "refs": ["http://www.genialis.com/jobs"],
                },
                {"url": "http://www.dictyexpress.org"},
            ]
        }
        validate_schema(instance, schema)

        instance = {"webpage": {"url": "http://www.dictyexpress.org"}}
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

    def test_groups(self):
        schema = [
            {
                "name": "test_group",
                "group": [
                    {
                        "name": "result_file",
                        "type": "basic:file:",
                        "validate_regex": r"^.*\.txt$",
                    },
                    {"name": "description", "type": "basic:string:", "required": True},
                ],
            }
        ]

        instance = {
            "test_group": {
                "result_file": {"file": "results.txt"},
                "description": "This are results",
            }
        }
        validate_schema(instance, schema)

        # wrong file extension
        instance = {
            "test_group": {
                "result_file": {"file": "results.tar.gz"},
                "description": "This are results",
            }
        }
        with self.assertRaisesRegex(
            ValidationError, "File name .* does not match regex"
        ):
            validate_schema(instance, schema)

        # wrong description type
        instance = {
            "test_group": {
                "result_file": {"file": "results.txt"},
                "description": 6,
            }
        }
        with self.assertRaisesRegex(ValidationError, "is not valid"):
            validate_schema(instance, schema)

        # missing description
        instance = {
            "test_group": {
                "result_file": {"file": "results.txt"},
            }
        }
        with self.assertRaisesRegex(ValidationError, '"description" not given'):
            validate_schema(instance, schema)



# pylint: disable=missing-docstring
import copy
import os
from pathlib import Path

from asgiref.sync import async_to_sync

from django.conf import settings
from django.test import override_settings

from guardian.shortcuts import assign_perm

from resolwe.flow.managers import manager
from resolwe.flow.managers.utils import disable_auto_calls
from resolwe.flow.models import (
    Collection,
    Data,
    DataDependency,
    DescriptorSchema,
    Process,
)
from resolwe.test import ProcessTestCase, TransactionTestCase

PROCESSES_DIR = os.path.join(os.path.dirname(__file__), "processes")


class TestManager(ProcessTestCase):
    def setUp(self):
        super().setUp()

        self.collection = Collection.objects.create(contributor=self.contributor)

        self._register_schemas(processes_paths=[PROCESSES_DIR])

    def test_create_data(self):
        """Test that manager is run when new object is created."""
        process = Process.objects.filter(slug="test-min").latest()
        data = Data.objects.create(
            name="Test data",
            contributor=self.contributor,
            process=process,
        )

        data.refresh_from_db()
        self.assertEqual(data.status, Data.STATUS_DONE)

    def test_spawned_process(self):
        """Test that manager is run for spawned processes and permissions are copied."""
        DescriptorSchema.objects.create(
            name="Test schema", slug="test-schema", contributor=self.contributor
        )
        spawned_process = Process.objects.filter(slug="test-save-file").latest()
        # Patch the process to create Entity, so its bahaviour can be tested.
        spawned_process.entity_type = "test-schema"
        spawned_process.entity_descriptor_schema = "test-schema"
        spawned_process.save()

        assign_perm("view_collection", self.user, self.collection)
        Data.objects.create(
            name="Test data",
            contributor=self.contributor,
            process=Process.objects.filter(slug="test-spawn-new").latest(),
            collection=self.collection,
        )

        # Created and spawned objects should be done.
        self.assertEqual(Data.objects.filter(status=Data.STATUS_DONE).count(), 2)

        # Check that permissions are inherited.
        child = Data.objects.last()
        self.assertTrue(self.user.has_perm("view_data", child))
        self.assertEqual(child.collection.pk, self.collection.pk)
        self.assertEqual(child.entity.collection.pk, self.collection.pk)

    def test_workflow(self):
        """Test that manager is run for workflows."""
        workflow = Process.objects.filter(slug="test-workflow-1").latest()
        data1 = Data.objects.create(
            name="Test data 1",
            contributor=self.contributor,
            process=workflow,
            input={"param1": "world"},
        )
        data2 = Data.objects.create(
            name="Test data 2",
            contributor=self.contributor,
            process=workflow,
            input={"param1": "foobar"},
        )

        # Created and spawned objects should be done.
        self.assertEqual(Data.objects.filter(status=Data.STATUS_DONE).count(), 6)

        # Check correct dependency type is created.
        self.assertEqual(
            {d.kind for d in data1.children_dependency.all()},
            {DataDependency.KIND_SUBPROCESS},
        )
        self.assertEqual(
            {d.kind for d in data2.children_dependency.all()},
            {DataDependency.KIND_SUBPROCESS},
        )

    def test_communicate(self):
        """Make sure communicate can serialize settings.

        When object of PosixPath is used in settings the call to communicate
        fails due to object not being JSON serializable by default.
        """
        flow_executor = copy.deepcopy(settings.FLOW_EXECUTOR)
        flow_executor["DATA_DIR"] = Path(flow_executor["DATA_DIR"])
        with override_settings(FLOW_EXECUTOR=flow_executor):
            process = Process.objects.filter(slug="test-min").latest()
            Data.objects.create(
                name="Test data",
                contributor=self.contributor,
                process=process,
            )

    def test_dependencies(self):
        """Test that manager handles dependencies correctly."""
        process_parent = Process.objects.filter(slug="test-dependency-parent").latest()
        process_child = Process.objects.filter(slug="test-dependency-child").latest()
        data_parent = Data.objects.create(
            name="Test parent", contributor=self.contributor, process=process_parent
        )
        data_child1 = Data.objects.create(
            name="Test child",
            contributor=self.contributor,
            process=process_child,
            input={},
        )
        data_child2 = Data.objects.create(
            name="Test child",
            contributor=self.contributor,
            process=process_child,
            input={"parent": data_parent.pk},
        )
        data_child3 = Data.objects.create(
            name="Test child",
            contributor=self.contributor,
            process=process_child,
            input={"parent": None},
        )

        data_parent.refresh_from_db()
        data_child1.refresh_from_db()
        data_child2.refresh_from_db()
        data_child3.refresh_from_db()

        # Check locks are created in manager.
        self.assertFalse(data_parent.access_logs.exists())
        self.assertFalse(data_child1.access_logs.exists())
        self.assertTrue(data_child2.access_logs.exists())
        self.assertFalse(data_child3.access_logs.exists())

        # Check that the data_parent location was locked.
        access_log = data_child2.access_logs.get()
        self.assertEqual(
            access_log.storage_location.file_storage.data.get().id, data_parent.id
        )
        # Check that the log is released.
        self.assertIsNotNone(access_log.started)
        self.assertIsNotNone(access_log.finished)

        # Check status.
        self.assertEqual(data_parent.status, Data.STATUS_DONE)
        self.assertEqual(data_child1.status, Data.STATUS_DONE)
        self.assertEqual(data_child2.status, Data.STATUS_DONE)
        self.assertEqual(data_child3.status, Data.STATUS_DONE)

    def test_process_notifications(self):
        process = Process.objects.filter(slug="test-process-notifications").latest()
        data = Data.objects.create(
            name="Test data",
            contributor=self.contributor,
            process=process,
        )

        data.refresh_from_db()

        self.assertEqual(len(data.process_info), 2)
        self.assertEqual(data.process_info[0], "abc")
        self.assertEqual(data.process_info[1][-5:], "xx...")

        self.assertEqual(len(data.process_warning), 1)
        self.assertEqual(data.process_warning[0][-5:], "yy...")

        self.assertEqual(len(data.process_error), 1)
        self.assertEqual(data.process_error[0][-5:], "zz...")


class TransactionTestManager(TransactionTestCase):
    @disable_auto_calls()
    def test_communicate(self):
        process = Process.objects.create(
            name="Input process",
            contributor=self.contributor,
            type="data:test:",
            input_schema=[
                {
                    "name": "input_data",
                    "type": "data:test:",
                    "required": False,
                },
            ],
        )

        data_1 = Data.objects.create(contributor=self.contributor, process=process)
        data_2 = Data.objects.create(
            contributor=self.contributor,
            process=process,
            input={"input_data": data_1.id},
        )
        Data.objects.create(contributor=self.contributor, process=process)
        Data.objects.create(contributor=self.contributor, process=process)

        self.assertEqual(Data.objects.filter(status=Data.STATUS_RESOLVING).count(), 4)

        # Allow unfinished data objects to exist when checking for execution
        # barrier condition in the dispatcher.
        async_to_sync(manager.communicate)(data_id=data_1.pk, run_sync=True)

        data_1.refresh_from_db()
        self.assertEqual(data_1.status, Data.STATUS_WAITING)
        self.assertEqual(Data.objects.filter(status=Data.STATUS_RESOLVING).count(), 3)

        data_1.status = Data.STATUS_DONE
        data_1.save()

        # Process object's children.
        async_to_sync(manager.communicate)(data_id=data_1.pk, run_sync=True)

        data_2.refresh_from_db()
        self.assertEqual(data_2.status, Data.STATUS_WAITING)
        self.assertEqual(Data.objects.filter(status=Data.STATUS_RESOLVING).count(), 2)

        # Process all objects.
        async_to_sync(manager.communicate)(run_sync=True)

        self.assertEqual(Data.objects.filter(status=Data.STATUS_RESOLVING).count(), 0)



# Generated by Django 2.2.2 on 2019-07-24 07:30

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0038_remove_m2m_fields"),
    ]

    operations = [
        migrations.AlterField(
            model_name="data",
            name="collection",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="data",
                to="flow.Collection",
            ),
        ),
        migrations.AlterField(
            model_name="data",
            name="entity",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="data",
                to="flow.Entity",
            ),
        ),
        migrations.AlterField(
            model_name="entity",
            name="collection",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="flow.Collection",
            ),
        ),
    ]



"""Resolwe models validation utils."""
import json
import re
from pathlib import Path

import jsonschema

from django.contrib.staticfiles import finders
from django.core.exceptions import ValidationError

from resolwe.flow.utils import dict_dot, iterate_dict, iterate_fields, iterate_schema


class DirtyError(ValidationError):
    """Error raised when required fields missing."""


def validation_schema(name):
    """Return json schema for json validation."""
    schemas = {
        "processor": "processSchema.json",
        "descriptor": "descriptorSchema.json",
        "field": "fieldSchema.json",
        "type": "typeSchema.json",
    }

    if name not in schemas:
        raise ValueError()

    field_schema_file = finders.find("flow/{}".format(schemas["field"]), all=True)[0]
    with open(field_schema_file, "r") as fn:
        field_schema = fn.read()

    if name == "field":
        return json.loads(field_schema.replace("{{PARENT}}", ""))

    schema_file = finders.find("flow/{}".format(schemas[name]), all=True)[0]
    with open(schema_file, "r") as fn:
        schema = fn.read()

    return json.loads(
        schema.replace("{{FIELD}}", field_schema).replace("{{PARENT}}", "/field")
    )


TYPE_SCHEMA = validation_schema("type")


def validate_schema(
    instance, schema, test_required=True, data_location=None, skip_missing_data=False
):
    """Check if DictField values are consistent with our data types.

    Perform basic JSON schema validation and our custom validations:

      * check that required fields are given (if `test_required` is set
        to ``True``)
      * check if ``basic:file:`` and ``list:basic:file`` fields match
        regex given in schema (only if ``validate_regex`` is defined in
        schema for coresponding fields) and exists (only if
        ``data_location`` is given)
      * check if directories referenced in ``basic:dir:`` and
        ``list:basic:dir``fields exist (only if ``data_location`` is
        given)
      * check that referenced ``Data`` objects (in ``data:<data_type>``
        and  ``list:data:<data_type>`` fields) exists and are of type
        ``<data_type>``
      * check that referenced ``Storage`` objects (in ``basic:json``
        fields) exists

    :param list instance: Instance to be validated
    :param list schema: Schema for validation
    :param bool test_required: Flag for testing if all required fields
        are present. It is usefule if validation is run before ``Data``
        object is finished and there are some field stil missing
        (default: ``False``)
    :param :class:`~resolwe.storage.models.FileStorage` data_location:
        data location used for checking if files and directories exist
        (default: ``None``)
    :param bool skip_missing_data: Don't raise an error if referenced
        ``Data`` object does not exist
    :rtype: None
    :raises ValidationError: if ``instance`` doesn't match schema
        defined in ``schema``

    """
    from ..storage import Storage  # Prevent circular import.

    path_prefix = None
    if data_location:
        path_prefix = Path(data_location.get_path())

    def validate_refs(field):
        """Validate reference paths."""
        for ref_filename in field.get("refs", []):
            ref_path: Path = path_prefix / ref_filename
            file_exists = ref_path.exists()
            if not file_exists:
                raise ValidationError(
                    "Path referenced in `refs` ({}) does not exist.".format(ref_path)
                )
            if not (ref_path.is_file() or ref_path.is_dir()):
                raise ValidationError(
                    "Path referenced in `refs` ({}) is neither a file or directory.".format(
                        ref_path
                    )
                )

    def validate_file(field, regex):
        """Validate file name (and check that it exists)."""
        filename = field["file"]
        if regex and not re.search(regex, filename):
            raise ValidationError(
                "File name {} does not match regex {}".format(filename, regex)
            )

        if path_prefix:
            path: Path = path_prefix / filename
            if not path.exists():
                raise ValidationError(
                    "Referenced path ({}) does not exist.".format(path)
                )
            if not path.is_file():
                raise ValidationError(
                    "Referenced path ({}) is not a file.".format(path)
                )

            validate_refs(field)

    def validate_dir(field):
        """Check that dirs and referenced files exists."""
        dirname = field["dir"]

        if path_prefix:
            path: Path = path_prefix / dirname
            if not path.exists():
                raise ValidationError(
                    "Referenced path ({}) does not exist.".format(path)
                )
            if not path.is_dir():
                raise ValidationError(
                    "Referenced path ({}) is not a directory.".format(path)
                )

            validate_refs(field)

    def validate_data(data_pk, type_):
        """Check that `Data` objects exist and is of right type."""
        from ..data import Data  # prevent circular import

        data_qs = Data.objects.filter(pk=data_pk).values("process__type")
        if not data_qs.exists():
            if skip_missing_data:
                return

            raise ValidationError(
                "Referenced `Data` object does not exist (id:{})".format(data_pk)
            )
        data = data_qs.first()
        if not data["process__type"].startswith(type_):
            raise ValidationError(
                "Data object of type `{}` is required, but type `{}` is given. "
                "(id:{})".format(type_, data["process__type"], data_pk)
            )

    def validate_range(value, interval, name):
        """Check that given value is inside the specified range."""
        if not interval:
            return

        if value < interval[0] or value > interval[1]:
            raise ValidationError(
                "Value of field '{}' is out of range. It should be between {} and {}.".format(
                    name, interval[0], interval[1]
                )
            )

    is_dirty = False
    dirty_fields = []
    for _schema, _fields, _ in iterate_schema(instance, schema):
        name = _schema["name"]
        is_required = _schema.get("required", True)

        if test_required and is_required and name not in _fields:
            is_dirty = True
            dirty_fields.append(name)

        if name in _fields:
            field = _fields[name]
            type_ = _schema.get("type", "")

            # Treat None as if the field is missing.
            if not is_required and field is None:
                continue

            try:
                jsonschema.validate([{"type": type_, "value": field}], TYPE_SCHEMA)
            except jsonschema.exceptions.ValidationError as ex:
                raise ValidationError(ex.message)

            choices = [choice["value"] for choice in _schema.get("choices", [])]
            allow_custom_choice = _schema.get("allow_custom_choice", False)
            if choices and not allow_custom_choice and field not in choices:
                raise ValidationError(
                    "Value of field '{}' must match one of predefined choices. "
                    "Current value: {}".format(name, field)
                )

            if type_ == "basic:file:":
                validate_file(field, _schema.get("validate_regex"))

            elif type_ == "list:basic:file:":
                for obj in field:
                    validate_file(obj, _schema.get("validate_regex"))

            elif type_ == "basic:dir:":
                validate_dir(field)

            elif type_ == "list:basic:dir:":
                for obj in field:
                    validate_dir(obj)

            elif (
                type_ == "basic:json:" and not Storage.objects.filter(pk=field).exists()
            ):
                raise ValidationError(
                    "Referenced `Storage` object does not exist (id:{})".format(field)
                )

            elif type_.startswith("data:"):
                validate_data(field, type_)

            elif type_.startswith("list:data:"):
                for data_id in field:
                    validate_data(data_id, type_[5:])  # remove `list:` from type

            elif type_ == "basic:integer:" or type_ == "basic:decimal:":
                validate_range(field, _schema.get("range"), name)

            elif type_ == "list:basic:integer:" or type_ == "list:basic:decimal:":
                for obj in field:
                    validate_range(obj, _schema.get("range"), name)

    try:
        # Check that schema definitions exist for all fields
        for _, _ in iterate_fields(instance, schema):
            pass
    except KeyError as ex:
        raise ValidationError(str(ex))

    if is_dirty:
        dirty_fields = ['"{}"'.format(field) for field in dirty_fields]
        raise DirtyError(
            "Required fields {} not given.".format(", ".join(dirty_fields))
        )


def validate_data_object(data, skip_missing_data=False):
    """Validate data object.

    Data object is validated only when worker is done with processing.
    """
    validate_schema(
        data.input, data.process.input_schema, skip_missing_data=skip_missing_data
    )
    validate_schema(data.output, data.process.output_schema)


def validate_process_subtype(supertype_name, supertype, subtype_name, subtype):
    """Perform process subtype validation.

    :param supertype_name: Supertype name
    :param supertype: Supertype schema
    :param subtype_name: Subtype name
    :param subtype: Subtype schema
    :return: A list of validation error strings
    """
    errors = []
    for item in supertype:
        # Ensure that the item exists in subtype and has the same schema.
        for subitem in subtype:
            if item["name"] != subitem["name"]:
                continue

            for key in set(item.keys()) | set(subitem.keys()):
                if key in ("label", "description"):
                    # Label and description can differ.
                    continue
                elif key == "required":
                    # A non-required item can be made required in subtype, but not the
                    # other way around.
                    item_required = item.get("required", True)
                    subitem_required = subitem.get("required", False)

                    if item_required and not subitem_required:
                        errors.append(
                            "Field '{}' is marked as required in '{}' and optional in '{}'.".format(
                                item["name"], supertype_name, subtype_name
                            )
                        )
                elif item.get(key, None) != subitem.get(key, None):
                    errors.append(
                        "Schema for field '{}' in type '{}' does not match supertype '{}'.".format(
                            item["name"], subtype_name, supertype_name
                        )
                    )

            break
        else:
            errors.append(
                "Schema for type '{}' is missing supertype '{}' field '{}'.".format(
                    subtype_name, supertype_name, item["name"]
                )
            )

    return errors


def validate_process_types(queryset=None):
    """Perform process type validation.

    :param queryset: Optional process queryset to validate
    :return: A list of validation error strings
    """
    if not queryset:
        from ..process import Process

        queryset = Process.objects.all()

    processes = {}
    for process in queryset:
        dict_dot(
            processes,
            process.type.replace(":", ".") + "__schema__",
            process.output_schema,
        )

    errors = []
    for path, key, value in iterate_dict(
        processes, exclude=lambda key, value: key == "__schema__"
    ):
        if "__schema__" not in value:
            continue

        # Validate with any parent types.
        for length in range(len(path), 0, -1):
            parent_type = ".".join(path[:length] + ["__schema__"])
            try:
                parent_schema = dict_dot(processes, parent_type)
            except KeyError:
                continue

            errors += validate_process_subtype(
                supertype_name=":".join(path[:length]),
                supertype=parent_schema,
                subtype_name=":".join(path + [key]),
                subtype=value["__schema__"],
            )

    return errors



"""Amazon S3 storage connector."""
import json
import logging
import mimetypes
import os
from pathlib import Path

import boto3
import botocore

from .baseconnector import BaseStorageConnector, validate_url, validate_urls

logger = logging.getLogger(__name__)


class AwsS3Connector(BaseStorageConnector):
    """Amazon S3 storage connector."""

    REQUIRED_SETTINGS = ["bucket", "credentials"]

    def __init__(self, config: dict, name: str):
        """Connector initialization."""
        super().__init__(config, name)
        self.bucket_name = config["bucket"]
        self.supported_hash = ["awss3etag"]
        self.hash_propery = {"awss3etag": "e_tag"}
        self.multipart_threshold = self.config.get(
            "multipart_threshold", self.CHUNK_SIZE
        )
        self.multipart_chunksize = self.config.get(
            "multipart_chunksize", self.CHUNK_SIZE
        )
        self.use_threads = True

        # Ensured by TLS protocol used for transport.
        self.get_ensures_data_integrity = True

        self._session = None
        self._client = None
        self._sts = None
        self._awss3 = None

    def _initialize(self):
        """Initializaton."""
        credentials = self.config["credentials"]
        with open(credentials) as f:
            settings = json.load(f)
        self.session = boto3.Session(
            aws_access_key_id=settings["AccessKeyId"],
            aws_secret_access_key=settings["SecretAccessKey"],
            aws_session_token=settings.get("SessionToken"),
            region_name=self.config.get("region_name"),
        )
        self.awss3 = self.session.resource("s3")

        self.client = self.session.client(
            "s3", config=botocore.client.Config(signature_version="s3v4")
        )
        self.client = self.session.client("s3")
        self.sts = self.session.client("sts")

    def __getattr__(self, name):
        """Lazy initialize some attributes."""
        requires_initialization = ["client", "awss3", "sts", "session"]
        if name not in requires_initialization:
            raise AttributeError()

        self._initialize()
        return getattr(self, name)

    @validate_url
    def push(self, stream, url, chunk_size=BaseStorageConnector.CHUNK_SIZE):
        """Push data from the stream to the given URL."""
        url = os.fspath(url)
        mime_type = mimetypes.guess_type(url)[0]
        extra_args = {} if mime_type is None else {"ContentType": mime_type}
        extra_args["Metadata"] = {"_upload_chunk_size": str(chunk_size)}
        self.client.upload_fileobj(
            stream,
            self.bucket_name,
            url,
            Config=self._get_transfer_config(chunk_size),
            ExtraArgs=extra_args,
        )

    @validate_urls
    @validate_url
    def delete(self, url, urls):
        """Remove objects."""
        # At most 1000 objects can be deleted at the same time.
        max_chunk = 1000
        bucket = self.awss3.Bucket(self.bucket_name)
        for i in range(0, len(urls), max_chunk):
            next_chunk = urls[i : i + max_chunk]
            objects = [
                {"Key": os.fspath(self.base_path / url / delete_url)}
                for delete_url in next_chunk
            ]
            bucket.delete_objects(Delete={"Objects": objects, "Quiet": True})

    @validate_url
    def get(self, url, stream, chunk_size=BaseStorageConnector.CHUNK_SIZE):
        """Get data from the given URL and write it into the given stream."""
        chunk_size = max(chunk_size, self.multipart_threshold)
        self.client.download_fileobj(
            self.bucket_name,
            os.fspath(url),
            stream,
            Config=self._get_transfer_config(chunk_size),
        )

    def _get_transfer_config(self, chunk_size=BaseStorageConnector.CHUNK_SIZE):
        """Get transfer config object."""
        chunk_size = max(chunk_size, self.multipart_threshold)
        return boto3.s3.transfer.TransferConfig(
            multipart_threshold=self.multipart_threshold,
            multipart_chunksize=chunk_size,
            use_threads=self.use_threads,
        )

    @validate_url
    def get_object_list(self, url):
        """Get a list of objects stored bellow the given URL."""
        url = os.path.join(url, "")
        paginator = self.client.get_paginator("list_objects_v2")
        kwargs = {"Bucket": self.bucket_name, "Prefix": url}
        ret = []
        for page in paginator.paginate(**kwargs):
            try:
                contents = page["Contents"]
            except KeyError:
                break
            for obj in contents:
                ret.append(Path(obj["Key"]).relative_to(url).as_posix())
        return ret

    @validate_url
    def get_hash(self, url, hash_type):
        """Get the hash of the given type for the given object."""
        resource = self.awss3.Object(self.bucket_name, os.fspath(url))
        try:
            if hash_type in self.hash_propery:
                prop = self.hash_propery[hash_type]
                return getattr(resource, prop).strip('"')
            else:
                return resource.metadata.get(hash_type)
        except botocore.exceptions.ClientError as error:
            if error.response["Error"]["Code"] == "404":
                return None
            else:
                # Something else has gone wrong.
                raise

    @validate_url
    def get_hashes(self, url, hash_types):
        """Get the hash of the given type for the given object."""
        resource = self.awss3.Object(self.bucket_name, os.fspath(url))
        hashes = dict()
        try:
            for hash_type in hash_types:
                if hash_type in self.hash_propery:
                    prop = self.hash_propery[hash_type]
                    hashes[hash_type] = getattr(resource, prop).strip('"')
                else:
                    hashes[hash_type] = resource.metadata.get(hash_type)
        except botocore.exceptions.ClientError as error:
            if error.response["Error"]["Code"] == "404":
                return None
            else:
                # Something else has gone wrong.
                raise
        return hashes

    @validate_url
    def exists(self, url):
        """Get if the object at the given URL exist."""
        try:
            self.awss3.Object(self.bucket_name, os.fspath(url)).load()
        except botocore.exceptions.ClientError as error:
            if error.response["Error"]["Code"] == "404":
                return False
            else:
                # Something else has gone wrong.
                raise
        else:
            return True

    @validate_url
    def set_hashes(self, url, hashes):
        """Set the  hashes for the given object."""
        # Changing metadata on existing objects in S3 is annoyingly hard.
        # See
        # https://boto3.amazonaws.com/v1/documentation/api/1.9.42/guide/s3.html
        # under managed copy for example.
        # If one uses copy_object method proposed by some solutions the e_tag
        # value on object can (and will) change. That causes error downloading
        # since hash check fails.
        url = os.fspath(url)
        head = self.client.head_object(Bucket=self.bucket_name, Key=url)
        content_type = head["ResponseMetadata"]["HTTPHeaders"]["content-type"]
        meta = head["Metadata"]
        chunk_size = int(meta["_upload_chunk_size"])
        hashes = {k: v for (k, v) in hashes.items() if k not in self.hash_propery}
        meta.update(hashes)
        copy_source = {
            "Bucket": self.bucket_name,
            "Key": url,
        }
        self.client.copy(
            copy_source,
            self.bucket_name,
            url,
            ExtraArgs={
                "Metadata": meta,
                "MetadataDirective": "REPLACE",
                "ContentType": content_type,
            },
            Config=self._get_transfer_config(chunk_size),
        )

    @property
    def base_path(self):
        """Get a base path for this connector."""
        return Path("")

    @validate_url
    def presigned_url(self, url, expiration=60, force_download=False):
        """Create a presigned URL.

        The URL is used to obtain temporary access to the object ar the
        given URL using only returned URL.

        :param expiration: expiration time of the link (in seconds), default
            is one minute.

        :param force_download: force download.

        :returns: URL that can be used to access object or None.
        """
        content_disposition = "attachment" if force_download else "inline"
        response = None
        try:
            response = self.client.generate_presigned_url(
                "get_object",
                Params={
                    "Bucket": self.bucket_name,
                    "Key": os.fspath(url),
                    "ResponseContentDisposition": content_disposition,
                },
                ExpiresIn=expiration,
            )
        except botocore.exceptions.ClientError:
            logger.exception("Error creating presigned URL")

        # The response contains the presigned URL
        return response



"""Custom permissions for Flow API."""
from django.contrib.auth.models import AnonymousUser
from django.http import Http404

from rest_framework import permissions


class ResolwePermissions(permissions.DjangoObjectPermissions):
    """Resolwe Permissions."""

    perms_map = {
        "GET": ["%(app_label)s.view_%(model_name)s"],
        "OPTIONS": ["%(app_label)s.view_%(model_name)s"],
        "HEAD": ["%(app_label)s.view_%(model_name)s"],
        "POST": ["%(app_label)s.edit_%(model_name)s"],
        "PUT": ["%(app_label)s.edit_%(model_name)s"],
        "PATCH": ["%(app_label)s.edit_%(model_name)s"],
        "DELETE": ["%(app_label)s.edit_%(model_name)s"],
    }

    def has_permission(self, request, view):
        """Return `True` as we don't use model level perms."""
        return True

    def has_object_permission(self, request, view, obj):
        """Check object permissions."""
        # admins can do anything
        if request.user.is_superuser:
            return True

        # `share` permission is required for editing permissions
        if "permissions" in view.action:
            self.perms_map["POST"] = ["%(app_label)s.share_%(model_name)s"]

        if view.action in ["add_data", "remove_data"]:
            self.perms_map["POST"] = ["%(app_label)s.add_%(model_name)s"]

        if hasattr(view, "get_queryset"):
            queryset = view.get_queryset()
        else:
            queryset = getattr(view, "queryset", None)

        assert queryset is not None, (
            "Cannot apply DjangoObjectPermissions on a view that "
            "does not set `.queryset` or have a `.get_queryset()` method."
        )

        model_cls = queryset.model
        user = request.user

        perms = self.get_required_object_permissions(request.method, model_cls)

        if not user.has_perms(perms, obj) and not AnonymousUser().has_perms(perms, obj):
            # If the user does not have permissions we need to determine if
            # they have read permissions to see 403, or not, and simply see
            # a 404 response.

            if request.method in permissions.SAFE_METHODS:
                # Read permissions already checked and failed, no need
                # to make another lookup.
                raise Http404

            read_perms = self.get_required_object_permissions("GET", model_cls)
            if not user.has_perms(read_perms, obj):
                raise Http404

            # Has read permissions.
            return False

        return True



"""Commands specific to kubernetes installation."""

from typing import TYPE_CHECKING

from resolwe.flow.executors.socket_utils import Message, Response
from resolwe.flow.utils import iterate_fields
from resolwe.storage.models import FileStorage, ReferencedPath

from .plugin import ListenerPlugin

if TYPE_CHECKING:
    from resolwe.flow.managers.listener.listener import Processor


class InitContainerPlugin(ListenerPlugin):
    """Handler methods for Init Container."""

    name = "Init container plugin"

    def handle_get_inputs_no_shared_storage(
        self, message: Message[int], manager: "Processor"
    ) -> Response:
        """Get a files belonging to input data objects.

        The format of the output is as follows:
        {
            base_url_1: (connector1, [list, of, ReferencedPath, instances]),
            bose_url_2: (connector2, [another, list, of, ReferencedPath, instances])
            ...
        }
        """
        output_data = {}

        # First get ids of data objecs which are inputs for data object we are
        # processing.
        input_data_ids = []
        for schema, fields in iterate_fields(
            manager.data.input, manager.data.process.input_schema
        ):
            type_ = schema["type"]
            if type_.startswith("data:") or type_.startswith("list:data:"):
                value = fields[schema["name"]]
                if isinstance(value, int):
                    input_data_ids.append(value)
                else:
                    input_data_ids += value

        for input_data_id in input_data_ids:
            file_storage = FileStorage.objects.get(data=input_data_id)
            location = file_storage.default_storage_location
            output_data[location.url] = (
                location.connector_name,
                list(
                    ReferencedPath.objects.filter(storage_locations=location).values()
                ),
            )

        manager._listener.communicator.suspend_heartbeat(manager.peer_identity)
        return message.respond_ok(output_data)



"""Python process runner utility."""
import argparse
import inspect
import logging
import os
import sys
from importlib import import_module
from pathlib import Path
from typing import Dict, Type

from .descriptor import ValidationError
from .models import Data
from .runtime import Process

# Id of the Data object we are processing.
DATA_ID = int(os.getenv("DATA_ID", "-1"))
logger = logging.getLogger(__name__)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run a Resolwe Python proces")
    parser.add_argument("filename", type=str, help="Python process filename to run")
    args = parser.parse_args()

    # Get the data object.
    data = Data(DATA_ID)

    # Switch to target directory to import the module.
    try:
        # Insert empty string as first path argument due to the issue
        # https://bugs.python.org/issue33053
        sys.path.insert(0, "")
        filename = Path(args.filename).resolve()
        start_dir = Path.cwd()
        os.chdir(filename.parent)
        module = import_module(filename.stem, __package__)
        os.chdir(start_dir)
    except (OSError, ImportError):
        logger.exception("Failed to load Python process from %s.", args.filename)
        sys.exit(1)

    # Mapping between the process slug and the class containing the process
    # definition.
    processes: Dict[str, Type[Process]] = {}
    for variable in dir(module):
        value = getattr(module, variable)
        if (
            value == Process
            or not inspect.isclass(value)
            or not issubclass(value, Process)
            or "_abstract" in value.__dict__
        ):
            continue
        processes[value._meta.metadata.slug] = value

    try:
        process = processes[data.process.slug]
    except KeyError:
        print("Found the following processes in module '{}':".format(args.module))
        print("")
        for slug, process in processes.items():
            print("  {} ({})".format(slug, getattr(process, "name")))
        print("")

        print("ERROR: Unable to find process '{}'.".format(data.process.slug))
        sys.exit(1)

    # Start the process.
    try:
        process(data).start()
    except ValidationError as error:
        print("ERROR: Output field validation failed: {}".format(error.args[0]))
        sys.exit(1)
